#include <engine/physics.h>
#include <engine/display.h>

/*
 * mass_draw:
 *	Draws the mass's debug members.
 */
mass_draw(p_mass)
struct Mass *p_mass;
{
	if(p_mass->is_debug)
	{
		DrawPrim(&p_mass->debug_pos);
		DrawPrim(&p_mass->debug_vel);
		DrawPrim(&p_mass->debug_force);
	}
}

/*
 * mass_init:
 *	Initializes the members of a Mass struct.
 */
mass_init(p_mass)
struct Mass *p_mass;
{
	p_mass->m = 1<<FIXED_PRECISION; /* 1kg */

	p_mass->is_debug = 0;

	setVector(&p_mass->pos, 0, 0, 0);
	setVector(&p_mass->vel, 0, 0, 0);
	setVector(&p_mass->force, 0, 0, 0);

	SetTile8(&p_mass->debug_pos);
	setRGB0(&p_mass->debug_pos, 255, 0, 0);

	setLineF2(&p_mass->debug_vel);
	setRGB0(&p_mass->debug_vel, 0, 255, 0);

	setLineF2(&p_mass->debug_force);
	setRGB0(&p_mass->debug_force, 0, 0, 255);
}

/*
 * mass_set_force:
 *	Overrides the force of a mass to a given
 *	vector.
 */
mass_set_force(p_mass, p_force)
struct Mass *p_mass;
VECTOR *p_force;
{
	p_mass->force.vx = p_force->vx;
	p_mass->force.vy = p_force->vy;
	p_mass->force.vz = p_force->vz;
}

/*
 * mass_add_force:
 *	Adds an external force to the net force
 *	of a given mass.
 */
mass_add_force(p_mass, p_force)
struct Mass *p_mass;
VECTOR *p_force;
{
	p_mass->force.vx += p_force->vx;
	p_mass->force.vy += p_force->vy;
	p_mass->force.vz += p_force->vz;
}

/*
 * mass_simulate:
 *	Applies the net force of the mass onto
 *	the velocity of the mass. The velocity
 *	is then applied to the position of the
 *	mass. i.e: F=ma
 */
mass_simulate(p_mass)
struct Mass *p_mass;
{
	int m = p_mass->m;

	VECTOR pos;
	VECTOR vel;
	VECTOR f;

	setVector(&pos, p_mass->pos.vx, p_mass->pos.vy, p_mass->pos.vz);
	setVector(&vel, p_mass->vel.vx, p_mass->vel.vy, p_mass->vel.vz);
	setVector(&f, p_mass->force.vx, p_mass->force.vy, p_mass->force.vz);

	/* FIXME: probably can round more efficiently */
	if( m%(1<<FIXED_PRECISION) >= (1<<FIXED_PRECISION)/2 )
	{
		m += (1<<FIXED_PRECISION);
	}

	m >>= FIXED_PRECISION;
	m = !m ? 1 : m;

	p_mass->vel.vx += (p_mass->force.vx / m)/METER;
	p_mass->vel.vy += (p_mass->force.vy / m)/METER;
	p_mass->vel.vz += (p_mass->force.vz / m)/METER;

	p_mass->pos.vx += p_mass->vel.vx;
	p_mass->pos.vy += p_mass->vel.vy;
	p_mass->pos.vz += p_mass->vel.vz;

	/* update debug members */
	setXY0(&p_mass->debug_pos, pos.vx>>FIXED_PRECISION, (-pos.vy)>>FIXED_PRECISION);

	#define INTENSITY	2

	setXY2(&p_mass->debug_vel, pos.vx>>FIXED_PRECISION, (-pos.vy)>>FIXED_PRECISION,
		((pos.vx + vel.vx)>>FIXED_PRECISION)*INTENSITY, (-((pos.vy + vel.vy)>>FIXED_PRECISION))*INTENSITY );


	setXY2(&p_mass->debug_force, pos.vx>>FIXED_PRECISION, (-pos.vy)>>FIXED_PRECISION,
		((pos.vx + f.vx)>>FIXED_PRECISION)*INTENSITY, (-((pos.vy + f.vy)>>FIXED_PRECISION))*INTENSITY );
}

/*
 * spring_solve:
 *	Adds force to a given spring via Hooke's
 *	law in three dimensions. The formula for
 *	the spring force is F = -k * (x - d).
 *	We also apply friction against the
 *	spring force itself.
 */
spring_solve(p_spring)
struct Spring *p_spring;
{
	int r;

	VECTOR v, u, f;

	setVector(&u, 0, 0, 0);
	setVector(&f, 0, 0, 0);

	setVector(&v,	p_spring->p_mass1->pos.vx - p_spring->p_mass2->pos.vx,
			p_spring->p_mass1->pos.vy - p_spring->p_mass2->pos.vy,
			p_spring->p_mass1->pos.vz - p_spring->p_mass2->pos.vz);

	r = psqrt(	((v.vx*v.vx)>>FIXED_PRECISION) +
			((v.vy*v.vy)>>FIXED_PRECISION) +
			((v.vz*v.vz)>>FIXED_PRECISION));


	/* determine unit vector */
	if( r != 0 )
	{
		if( r>>FIXED_PRECISION != 0 )
		{
			setVector(&u,	v.vx / (r>>FIXED_PRECISION),
					v.vy / (r>>FIXED_PRECISION),
					v.vz / (r>>FIXED_PRECISION));
		}
	}

	/* add spring force */
	f.vx += (((u.vx * (r - p_spring->length))>>FIXED_PRECISION) * -p_spring->k)>>FIXED_PRECISION;
	f.vy += (((u.vy * (r - p_spring->length))>>FIXED_PRECISION) * -p_spring->k)>>FIXED_PRECISION;
	f.vz += (((u.vz * (r - p_spring->length))>>FIXED_PRECISION) * -p_spring->k)>>FIXED_PRECISION;

	/* add spring friction force */
	f.vx += (-(p_spring->p_mass1->vel.vx - p_spring->p_mass2->vel.vx) * p_spring->friction)>>FIXED_PRECISION;
	f.vy += (-(p_spring->p_mass1->vel.vy - p_spring->p_mass2->vel.vy) * p_spring->friction)>>FIXED_PRECISION;
	f.vz += (-(p_spring->p_mass1->vel.vz - p_spring->p_mass2->vel.vz) * p_spring->friction)>>FIXED_PRECISION;

	/* apply the forces to the masses themselves */
	mass_add_force(p_spring->p_mass1, &f);

	f.vx = -f.vx;
	f.vy = -f.vy;
	f.vz = -f.vz;

	mass_add_force(p_spring->p_mass2, &f);
}

/*
 * rope_init:
 *	Initializes common members of a Rope struct.
 *	Values should be provided before calling this
 *	function, otherwise you'll get garbage.
 */
rope_init(p_rope)
struct Rope *p_rope;
{
	int i;

	/* rope's primitives initialization */
	for(i = 0; i < ROPE_NMASSES-1; i++)
	{
		setLineF2(&p_rope->prim[i]);
		setRGB0(&p_rope->prim[i], 255, 128, 128);
	}

	/* initialize rope's masses */
	for(i = 0; i < ROPE_NMASSES; i++)
	{
		mass_init(&p_rope->masses[i]);
		p_rope->masses[i].is_debug = 1;
		p_rope->masses[i].m = p_rope->mass_m;
		setVector( &p_rope->masses[i].pos, p_rope->origin_pos.vx + i*p_rope->spring_length, p_rope->origin_pos.vy, p_rope->origin_pos.vz + i*p_rope->spring_length );
		setVector( &p_rope->masses[i].vel, 0, 0, 0 );
		setVector( &p_rope->masses[i].force, 0, 0, 0 );
	}

	/* initialize rope's springs */
	for(i = 0; i < ROPE_NMASSES-1; i++)
	{
		p_rope->springs[i].p_mass1	= &p_rope->masses[i];
		p_rope->springs[i].p_mass2	= &p_rope->masses[i+1];
		p_rope->springs[i].k		= p_rope->spring_k;
		p_rope->springs[i].length	= p_rope->spring_length;
		p_rope->springs[i].friction	= p_rope->spring_friction;
	}
}

/*
 * rope_solve:
 *	Adds multiple forces upon all masses of a
 *	given rope. This includes gravitational force,
 *	air frction, spring friction, and the spring force.
 */
rope_solve(p_rope)
struct Rope *p_rope;
{
	int i;
	VECTOR v, u, f;

	for(i = 0; i < ROPE_NMASSES-1; i++)
	{
		spring_solve(&p_rope->springs[i]);
	}

	/* solve the rope last */
	for(i = 1; i < ROPE_NMASSES; i++)
	{
		VECTOR tmp = {0, 0, 0};

		/* add gravitational force */
		tmp.vy = (p_rope->gravity.vy * p_rope->masses[i].m)>>FIXED_PRECISION;
		mass_add_force(&p_rope->masses[i], &tmp);

		/* add air friction force */
		tmp.vx = (-p_rope->masses[i].vel.vx * p_rope->air_friction)>>FIXED_PRECISION;
		tmp.vy = (-p_rope->masses[i].vel.vy * p_rope->air_friction)>>FIXED_PRECISION;
		tmp.vz = (-p_rope->masses[i].vel.vz * p_rope->air_friction)>>FIXED_PRECISION;
		mass_add_force(&p_rope->masses[i], &tmp);

		/* get the velocity */
		v.vy = p_rope->masses[i].vel.vy;
		v.vx = 0;
		v.vz = 0;

		/* FIXME: if working with 2d, it should be (p_rope->masses[i].pos.vy + p_rope->masses[i].vel.vy < p_rope->ground_height) */
		/* if the mass is heading towards the ground, absorb energy */
		if(p_rope->masses[i].pos.vy + p_rope->masses[i].vel.vy > p_rope->ground_height)
		{
			p_rope->masses[i].vel.vy = p_rope->ground_height - (p_rope->masses[i].pos.vy + p_rope->masses[i].vel.vy);
		}
	}
}

/*
 * rope_simulate:
 *	Applies the net force upon each mass and spring
 *	of the rope.
 */
rope_simulate(p_rope)
struct Rope *p_rope;
{
	int i;

	for(i = 0; i < ROPE_NMASSES; i++)
	{
		p_rope->origin_pos.vx += p_rope->origin_vel.vx;
		p_rope->origin_pos.vy += p_rope->origin_vel.vy;
		p_rope->origin_pos.vz += p_rope->origin_vel.vz;

		/* FIXME: if working with 2d, it should be (p_rope->origin_pos.vy > p_rope->ground_height) */
		if(p_rope->origin_pos.vy > p_rope->ground_height)
		{
			p_rope->origin_pos.vy = p_rope->ground_height;
		}

		setXY0(&p_rope->masses[0].debug_pos, p_rope->masses[0].pos.vx>>FIXED_PRECISION,
			(-p_rope->masses[0].pos.vy)>>FIXED_PRECISION);

		p_rope->springs[0].p_mass1->pos.vx = p_rope->origin_pos.vx;
		p_rope->springs[0].p_mass1->pos.vy = p_rope->origin_pos.vy;
		p_rope->springs[0].p_mass1->pos.vz = p_rope->origin_pos.vz;

		if(i != 0)
		{
			mass_simulate(&p_rope->masses[i]);
		}
	}
}

/*
 * rope_draw:
 *	Draws the springs of the rope.
 */
rope_draw(p_rope, p_ot, p_global_pos)
struct Rope *p_rope;
unsigned long *p_ot;
SVECTOR *p_global_pos;
{
	int i;

	SVECTOR		dmy0, dmy1;
	MATRIX		rottrans;
	long		p, otz, flg, dmy_long;

	PushMatrix();

	RotMatrix_gte(&p_rope->rotation, &rottrans);
	TransMatrix(&rottrans, &p_rope->translation);

	SetRotMatrix(&rottrans);
	SetTransMatrix(&rottrans);

	/* update positions of LINE_F2s and draw */
	for(i = 0; i < ROPE_NMASSES-1; i++)
	{
		setVector(&dmy0, (p_rope->masses[i].pos.vx>>FIXED_PRECISION) + p_global_pos->vx,
			(p_rope->masses[i].pos.vy>>FIXED_PRECISION) + p_global_pos->vy,
			(p_rope->masses[i].pos.vz>>FIXED_PRECISION) + p_global_pos->vz);

		setVector(&dmy1, (p_rope->masses[i+1].pos.vx>>FIXED_PRECISION) + p_global_pos->vx,
			(p_rope->masses[i+1].pos.vy>>FIXED_PRECISION) + p_global_pos->vy,
			(p_rope->masses[i+1].pos.vz>>FIXED_PRECISION) + p_global_pos->vz);

		otz = RotAverage3(&dmy0, &dmy1, &dmy1,
			(long *)(&p_rope->prim[i].x0), (long *)(&p_rope->prim[i].x1),
			&dmy_long, &p, &flg);

		if(otz > MIN_OTZ && otz < game_otsize)
		{
			AddPrim(p_ot+otz, &p_rope->prim[i]);
		}

		DrawPrim(&p_rope->prim[i]);
	}

	PopMatrix();
}

